@@ -1,5 +1,5 @@
 /* eslint-disable @typescript-eslint/no-var-requires */
-import {
+import type {
   IWebpackCLI,
   WebpackCLICommandOption,
   WebpackCLIBuiltInOption,
@@ -43,24 +43,28 @@ import {
   Problem,
 } from "./types";
 
-import webpackMerge from "webpack-merge";
-import webpack from "webpack";
-import { Compiler, MultiCompiler, WebpackError, StatsOptions } from "webpack";
-import { stringifyStream } from "@discoveryjs/json-ext";
-import { Help, ParseOptions } from "commander";
+import type webpackMerge from "webpack-merge";
+import type webpack from "webpack";
+import { type Compiler, type MultiCompiler, type WebpackError, type StatsOptions } from "webpack";
+import { type stringifyStream } from "@discoveryjs/json-ext";
+import { type Help, type ParseOptions } from "commander";
 
-import { CLIPlugin as CLIPluginClass } from "./plugins/CLIPlugin";
+import { type CLIPlugin as CLIPluginClass } from "./plugins/cli-plugin";
 
 const fs = require("fs");
 const path = require("path");
 const { pathToFileURL } = require("url");
 const util = require("util");
-
 const { program, Option } = require("commander");
-const envinfo = require("envinfo");
 
-const WEBPACK_PACKAGE = process.env.WEBPACK_PACKAGE || "webpack";
-const WEBPACK_DEV_SERVER_PACKAGE = process.env.WEBPACK_DEV_SERVER_PACKAGE || "webpack-dev-server";
+const WEBPACK_PACKAGE_IS_CUSTOM = !!process.env.WEBPACK_PACKAGE;
+const WEBPACK_PACKAGE = WEBPACK_PACKAGE_IS_CUSTOM
+  ? (process.env.WEBPACK_PACKAGE as string)
+  : "webpack";
+const WEBPACK_DEV_SERVER_PACKAGE_IS_CUSTOM = !!process.env.WEBPACK_DEV_SERVER_PACKAGE;
+const WEBPACK_DEV_SERVER_PACKAGE = WEBPACK_DEV_SERVER_PACKAGE_IS_CUSTOM
+  ? (process.env.WEBPACK_DEV_SERVER_PACKAGE as string)
+  : "webpack-dev-server";
 
 interface Information {
   Binaries?: string[];
@@ -324,37 +328,83 @@ class WebpackCLI implements IWebpackCLI {
     process.exit(2);
   }
 
-  async tryRequireThenImport<T>(module: ModuleName, handleError = true): Promise<T> {
+  async tryRequireThenImport<T>(
+    module: ModuleName,
+    handleError = true,
+    moduleType: "unknown" | "commonjs" | "esm" = "unknown",
+  ): Promise<T> {
     let result;
 
-    try {
-      result = require(module);
-    } catch (error) {
-      const dynamicImportLoader: null | DynamicImport<T> =
-        require("./utils/dynamic-import-loader")();
-      if (
-        ((error as ImportLoaderError).code === "ERR_REQUIRE_ESM" ||
-          process.env.WEBPACK_CLI_FORCE_LOAD_ESM_CONFIG) &&
-        pathToFileURL &&
-        dynamicImportLoader
-      ) {
-        const urlForConfig = pathToFileURL(module);
+    switch (moduleType) {
+      case "unknown": {
+        try {
+          result = require(module);
+        } catch (error) {
+          const dynamicImportLoader: null | DynamicImport<T> =
+            require("./utils/dynamic-import-loader")();
+          if (
+            ((error as ImportLoaderError).code === "ERR_REQUIRE_ESM" ||
+              process.env.WEBPACK_CLI_FORCE_LOAD_ESM_CONFIG) &&
+            pathToFileURL &&
+            dynamicImportLoader
+          ) {
+            const urlForConfig = pathToFileURL(module);
 
-        result = await dynamicImportLoader(urlForConfig);
-        result = result.default;
+            result = await dynamicImportLoader(urlForConfig);
+            result = result.default;
 
-        return result;
+            return result;
+          }
+
+          if (handleError) {
+            this.logger.error(error);
+            process.exit(2);
+          } else {
+            throw error;
+          }
+        }
+        break;
       }
+      case "commonjs": {
+        try {
+          result = require(module);
+        } catch (error) {
+          if (handleError) {
+            this.logger.error(error);
+            process.exit(2);
+          } else {
+            throw error;
+          }
+        }
+        break;
+      }
+      case "esm": {
+        try {
+          const dynamicImportLoader: null | DynamicImport<T> =
+            require("./utils/dynamic-import-loader")();
 
-      if (handleError) {
-        this.logger.error(error);
-        process.exit(2);
-      } else {
-        throw error;
+          if (pathToFileURL && dynamicImportLoader) {
+            const urlForConfig = pathToFileURL(module);
+
+            result = await dynamicImportLoader(urlForConfig);
+            result = result.default;
+
+            return result;
+          }
+        } catch (error) {
+          if (handleError) {
+            this.logger.error(error);
+            process.exit(2);
+          } else {
+            throw error;
+          }
+        }
+
+        break;
       }
     }
 
-    // For babel/typescript
+    // For babel and other, only commonjs
     if (result && typeof result === "object" && "default" in result) {
       result = result.default || {};
     }
@@ -390,6 +440,7 @@ class WebpackCLI implements IWebpackCLI {
           },
         ],
         description: "To get the output in a specified format ( accept json or markdown )",
+        helpLevel: "minimum",
       },
       {
         name: "additional-package",
@@ -397,6 +448,7 @@ class WebpackCLI implements IWebpackCLI {
         configs: [{ type: "string" }],
         multiple: true,
         description: "Adds additional packages to the output",
+        helpLevel: "minimum",
       },
     ];
   }
@@ -449,6 +501,8 @@ class WebpackCLI implements IWebpackCLI {
 
     defaultInformation.npmPackages = `{${defaultPackages.map((item) => `*${item}*`).join(",")}}`;
 
+    const envinfo = await this.tryRequireThenImport<typeof import("envinfo")>("envinfo", false);
+
     let info = await envinfo.run(defaultInformation, envinfoConfig);
 
     info = info.replace(/npmPackages/g, "Packages");
@@ -473,7 +527,6 @@ class WebpackCLI implements IWebpackCLI {
     }
 
     const command = this.program.command(commandOptions.name, {
-      noHelp: commandOptions.noHelp,
       hidden: commandOptions.hidden,
       isDefault: commandOptions.isDefault,
     }) as WebpackCLICommand;
@@ -519,12 +572,12 @@ class WebpackCLI implements IWebpackCLI {
         let skipInstallation = false;
 
         // Allow to use `./path/to/webpack.js` outside `node_modules`
-        if (dependency === WEBPACK_PACKAGE && fs.existsSync(WEBPACK_PACKAGE)) {
+        if (dependency === WEBPACK_PACKAGE && WEBPACK_PACKAGE_IS_CUSTOM) {
           skipInstallation = true;
         }
 
         // Allow to use `./path/to/webpack-dev-server.js` outside `node_modules`
-        if (dependency === WEBPACK_DEV_SERVER_PACKAGE && fs.existsSync(WEBPACK_PACKAGE)) {
+        if (dependency === WEBPACK_DEV_SERVER_PACKAGE && WEBPACK_DEV_SERVER_PACKAGE_IS_CUSTOM) {
           skipInstallation = true;
         }
 
@@ -560,9 +613,9 @@ class WebpackCLI implements IWebpackCLI {
         }
       }
 
-      options.forEach((optionForCommand) => {
-        this.makeOption(command, optionForCommand);
-      });
+      for (const option of options) {
+        this.makeOption(command, option);
+      }
     }
 
     command.action(action);
@@ -584,7 +637,7 @@ class WebpackCLI implements IWebpackCLI {
       let negatedDescription;
       const mainOptionType: WebpackCLIMainOption["type"] = new Set();
 
-      option.configs.forEach((config) => {
+      for (const config of option.configs) {
         switch (config.type) {
           case "reset":
             mainOptionType.add(Boolean);
@@ -608,7 +661,7 @@ class WebpackCLI implements IWebpackCLI {
           case "enum": {
             let hasFalseEnum = false;
 
-            const enumTypes = (config.values || []).map((value) => {
+            for (const value of config.values || []) {
               switch (typeof value) {
                 case "string":
                   mainOptionType.add(String);
@@ -625,20 +678,19 @@ class WebpackCLI implements IWebpackCLI {
                   mainOptionType.add(Boolean);
                   break;
               }
-            });
+            }
 
             if (!needNegativeOption) {
               needNegativeOption = hasFalseEnum;
               negatedDescription = config.negatedDescription;
             }
-
-            return enumTypes;
           }
         }
-      });
+      }
 
       mainOption = {
         flags: option.alias ? `-${option.alias}, --${option.name}` : `--${option.name}`,
+        valueName: option.valueName || "value",
         description: option.description || "",
         type: mainOptionType,
         multiple: option.multiple,
@@ -655,6 +707,7 @@ class WebpackCLI implements IWebpackCLI {
     } else {
       mainOption = {
         flags: option.alias ? `-${option.alias}, --${option.name}` : `--${option.name}`,
+        valueName: option.valueName || "value",
         description: option.description || "",
         type: option.type
           ? new Set(Array.isArray(option.type) ? option.type : [option.type])
@@ -674,9 +727,13 @@ class WebpackCLI implements IWebpackCLI {
     }
 
     if (mainOption.type.size > 1 && mainOption.type.has(Boolean)) {
-      mainOption.flags = `${mainOption.flags} [value${mainOption.multiple ? "..." : ""}]`;
+      mainOption.flags = `${mainOption.flags} [${mainOption.valueName || "value"}${
+        mainOption.multiple ? "..." : ""
+      }]`;
     } else if (mainOption.type.size > 0 && !mainOption.type.has(Boolean)) {
-      mainOption.flags = `${mainOption.flags} <value${mainOption.multiple ? "..." : ""}>`;
+      mainOption.flags = `${mainOption.flags} <${mainOption.valueName || "value"}${
+        mainOption.multiple ? "..." : ""
+      }>`;
     }
 
     if (mainOption.type.size === 1) {
@@ -809,7 +866,10 @@ class WebpackCLI implements IWebpackCLI {
           },
         ],
         multiple: true,
-        description: "Provide path to a webpack configuration file e.g. ./webpack.config.js.",
+        valueName: "pathToConfigFile",
+        description:
+          'Provide path to one or more webpack configuration files to process, e.g. "./webpack.config.js".',
+        helpLevel: "minimum",
       },
       {
         name: "config-name",
@@ -819,7 +879,10 @@ class WebpackCLI implements IWebpackCLI {
           },
         ],
         multiple: true,
-        description: "Name of the configuration to use.",
+        valueName: "name",
+        description:
+          "Name(s) of particular configuration(s) to use if configuration file exports an array of multiple configurations.",
+        helpLevel: "minimum",
       },
       {
         name: "merge",
@@ -831,6 +894,7 @@ class WebpackCLI implements IWebpackCLI {
           },
         ],
         description: "Merge two or more configurations using 'webpack-merge'.",
+        helpLevel: "minimum",
       },
       {
         name: "disable-interpret",
@@ -841,6 +905,7 @@ class WebpackCLI implements IWebpackCLI {
           },
         ],
         description: "Disable interpret for loading the config file.",
+        helpLevel: "minimum",
       },
       // Complex configs
       {
@@ -887,7 +952,9 @@ class WebpackCLI implements IWebpackCLI {
           return previous;
         },
         multiple: true,
-        description: "Environment passed to the configuration when it is a function.",
+        description:
+          'Environment variables passed to the configuration when it is a function, e.g. "myvar" or "myvar=myval".',
+        helpLevel: "minimum",
       },
       {
         name: "node-env",
@@ -898,6 +965,7 @@ class WebpackCLI implements IWebpackCLI {
         ],
         multiple: false,
         description: "Sets process.env.NODE_ENV to the specified value.",
+        helpLevel: "minimum",
       },
       {
         name: "define-process-env-node-env",
@@ -908,7 +976,8 @@ class WebpackCLI implements IWebpackCLI {
         ],
         multiple: false,
         description:
-          "Sets process.env.NODE_ENV to the specified value. (Currently an alias for `--node-env`)",
+          "Sets process.env.NODE_ENV to the specified value. (Currently an alias for `--node-env`).",
+        helpLevel: "verbose",
       },
 
       // Adding more plugins
@@ -922,6 +991,7 @@ class WebpackCLI implements IWebpackCLI {
         ],
         multiple: false,
         description: "It invokes webpack-bundle-analyzer plugin to get bundle information.",
+        helpLevel: "minimum",
       },
       {
         name: "progress",
@@ -935,6 +1005,7 @@ class WebpackCLI implements IWebpackCLI {
           },
         ],
         description: "Print compilation progress during build.",
+        helpLevel: "minimum",
       },
 
       // Output options
@@ -950,7 +1021,9 @@ class WebpackCLI implements IWebpackCLI {
           },
         ],
         alias: "j",
+        valueName: "pathToJsonFile",
         description: "Prints result as JSON or store it in a file.",
+        helpLevel: "minimum",
       },
       {
         name: "fail-on-warnings",
@@ -960,12 +1033,26 @@ class WebpackCLI implements IWebpackCLI {
             values: [true],
           },
         ],
-        description: "Stop webpack-cli process with non-zero exit code on warnings from webpack",
+        description: "Stop webpack-cli process with non-zero exit code on warnings from webpack.",
+        helpLevel: "minimum",
+      },
+      // TODO remove this in the next major release, because not all webpack versions have this flag in CLI options
+      {
+        name: "extends",
+        alias: "e",
+        configs: [
+          {
+            type: "string",
+          },
+        ],
+        multiple: true,
+        description:
+          "Path to the configuration to be extended (only works when using webpack-cli).",
+        helpLevel: "minimum",
       },
     ];
 
     const minimumHelpFlags = [
-      ...builtInFlags.map((flag) => flag.name),
       "mode",
       "watch",
       "watch-options-stdin",
@@ -975,41 +1062,23 @@ class WebpackCLI implements IWebpackCLI {
       "target",
       "name",
       "output-path",
+      "extends",
     ];
 
     // Extract all the flags being exported from core.
     // A list of cli flags generated by core can be found here https://github.com/webpack/webpack/blob/main/test/__snapshots__/Cli.basictest.js.snap
-    const coreArguments = Object.entries(this.webpack.cli.getArguments()).map(([flag, meta]) => {
-      const inBuiltIn = builtInFlags.find((builtInFlag) => builtInFlag.name === flag);
-
-      if (inBuiltIn) {
-        return {
-          ...meta,
-          // @ts-expect-error this might be overwritten
-          name: flag,
-          group: "core",
-          ...inBuiltIn,
-          configs: meta.configs || [],
-        };
-      }
-
-      return { ...meta, name: flag, group: "core" };
-    });
-
-    const options: WebpackCLIBuiltInOption[] = ([] as WebpackCLIBuiltInFlag[])
-      .concat(
-        builtInFlags.filter(
-          (builtInFlag) =>
-            !coreArguments.find((coreArgument) => builtInFlag.name === coreArgument.name),
-        ),
-      )
-      .concat(coreArguments)
-      .map((option): WebpackCLIBuiltInOption => {
-        (option as WebpackCLIBuiltInOption).helpLevel = minimumHelpFlags.includes(option.name)
-          ? "minimum"
-          : "verbose";
-        return option as WebpackCLIBuiltInOption;
-      });
+    const options = builtInFlags.concat(
+      Object.entries(this.webpack.cli.getArguments()).map<WebpackCLIBuiltInOption>(
+        ([name, meta]) => {
+          return {
+            ...meta,
+            name,
+            group: "core",
+            helpLevel: minimumHelpFlags.includes(name) ? "minimum" : "verbose",
+          };
+        },
+      ),
+    );
 
     this.builtInOptionsCache = options;
 
@@ -1139,9 +1208,7 @@ class WebpackCLI implements IWebpackCLI {
           async () => {
             this.webpack = await this.loadWebpack();
 
-            return isWatchCommandUsed
-              ? this.getBuiltInOptions().filter((option) => option.name !== "watch")
-              : this.getBuiltInOptions();
+            return this.getBuiltInOptions();
           },
           async (entries, options) => {
             if (entries.length > 0) {
@@ -1262,11 +1329,11 @@ class WebpackCLI implements IWebpackCLI {
 
             const levenshtein = require("fastest-levenshtein");
 
-            (command as WebpackCLICommand).options.forEach((option) => {
+            for (const option of (command as WebpackCLICommand).options) {
               if (!option.hidden && levenshtein.distance(name, option.long?.slice(2)) < 3) {
                 this.logger.error(`Did you mean '--${option.name()}'?`);
               }
-            });
+            }
           }
         }
       }
@@ -1352,7 +1419,7 @@ class WebpackCLI implements IWebpackCLI {
           // Support multiple aliases
           subcommandTerm: (command: WebpackCLICommand) => {
             const humanReadableArgumentName = (argument: WebpackCLICommandOption) => {
-              const nameOutput = argument.name() + (argument.variadic === true ? "..." : "");
+              const nameOutput = argument.name() + (argument.variadic ? "..." : "");
 
               return argument.required ? "<" + nameOutput + ">" : "[" + nameOutput + "]";
             };
@@ -1372,6 +1439,14 @@ class WebpackCLI implements IWebpackCLI {
                 return false;
               }
 
+              // Hide `--watch` option when developer use `webpack watch --help`
+              if (
+                (options[0] === "w" || options[0] === "watch") &&
+                (option.name() === "watch" || option.name() === "no-watch")
+              ) {
+                return false;
+              }
+
               switch (option.helpLevel) {
                 case "verbose":
                   return isVerbose;
@@ -1726,8 +1801,14 @@ class WebpackCLI implements IWebpackCLI {
 
     const interpret = require("interpret");
     const loadConfigByPath = async (configPath: string, argv: Argv = {}) => {
-      const ext = path.extname(configPath);
-      const interpreted = Object.keys(interpret.jsVariants).find((variant) => variant === ext);
+      const ext = path.extname(configPath).toLowerCase();
+      let interpreted = Object.keys(interpret.jsVariants).find((variant) => variant === ext);
+      // Fallback `.cts` to `.ts`
+      // TODO implement good `.mts` support after https://github.com/gulpjs/rechoir/issues/43
+      // For ESM and `.mts` you need to use: 'NODE_OPTIONS="--loader ts-node/esm" webpack-cli --config ./webpack.config.mts'
+      if (!interpreted && /\.cts$/.test(ext)) {
+        interpreted = interpret.jsVariants[".ts"];
+      }
 
       if (interpreted && !disableInterpret) {
         const rechoir: Rechoir = require("rechoir");
@@ -1738,9 +1819,9 @@ class WebpackCLI implements IWebpackCLI {
           if ((error as RechoirError)?.failures) {
             this.logger.error(`Unable load '${configPath}'`);
             this.logger.error((error as RechoirError).message);
-            (error as RechoirError).failures.forEach((failure) => {
+            for (const failure of (error as RechoirError).failures) {
               this.logger.error(failure.error.message);
-            });
+            }
             this.logger.error("Please install one of them");
             process.exit(2);
           }
@@ -1754,10 +1835,24 @@ class WebpackCLI implements IWebpackCLI {
 
       type LoadConfigOption = PotentialPromise<WebpackConfiguration>;
 
+      let moduleType: "unknown" | "commonjs" | "esm" = "unknown";
+
+      switch (ext) {
+        case ".cjs":
+        case ".cts":
+          moduleType = "commonjs";
+          break;
+        case ".mjs":
+        case ".mts":
+          moduleType = "esm";
+          break;
+      }
+
       try {
         options = await this.tryRequireThenImport<LoadConfigOption | LoadConfigOption[]>(
           configPath,
           false,
+          moduleType,
         );
         // @ts-expect-error error type assertion
       } catch (error: Error) {
@@ -1772,6 +1867,11 @@ class WebpackCLI implements IWebpackCLI {
         process.exit(2);
       }
 
+      if (!options) {
+        this.logger.error(`Failed to load '${configPath}' config. Unable to find default export.`);
+        process.exit(2);
+      }
+
       if (Array.isArray(options)) {
         // reassign the value to assert type
         const optionsArray: ConfigOptions[] = options;
@@ -1842,45 +1942,47 @@ class WebpackCLI implements IWebpackCLI {
           }
 
           if (isArray) {
-            (loadedConfig.options as ConfigOptions[]).forEach((item) => {
+            for (const item of loadedConfig.options as ConfigOptions[]) {
               (config.options as ConfigOptions[]).push(item);
-            });
+            }
           } else {
             config.options.push(loadedConfig.options as WebpackConfiguration);
           }
         }
 
         if (isArray) {
-          (loadedConfig.options as ConfigOptions[]).forEach((options) => {
-            config.path.set(options, loadedConfig.path);
-          });
+          for (const options of loadedConfig.options as ConfigOptions[]) {
+            config.path.set(options, [loadedConfig.path]);
+          }
         } else {
-          config.path.set(loadedConfig.options, loadedConfig.path);
+          config.path.set(loadedConfig.options, [loadedConfig.path]);
         }
       });
 
       config.options = config.options.length === 1 ? config.options[0] : config.options;
     } else {
+      // TODO ".mts" is not supported by `interpret`, need to add it
+      // Prioritize popular extensions first to avoid unnecessary fs calls
+      const extensions = [
+        ".js",
+        ".mjs",
+        ".cjs",
+        ".ts",
+        ".cts",
+        ".mts",
+        ...Object.keys(interpret.extensions),
+      ];
       // Order defines the priority, in decreasing order
-      const defaultConfigFiles = [
-        "webpack.config",
-        ".webpack/webpack.config",
-        ".webpack/webpackfile",
-      ]
-        .map((filename) =>
-          // Since .cjs is not available on interpret side add it manually to default config extension list
-          [...Object.keys(interpret.extensions), ".cjs"].map((ext) => ({
-            path: path.resolve(filename + ext),
-            ext: ext,
-            module: interpret.extensions[ext],
-          })),
-        )
-        .reduce((accumulator, currentValue) => accumulator.concat(currentValue), []);
+      const defaultConfigFiles = new Set(
+        ["webpack.config", ".webpack/webpack.config", ".webpack/webpackfile"].flatMap((filename) =>
+          extensions.map((ext) => path.resolve(filename + ext)),
+        ),
+      );
 
       let foundDefaultConfigFile;
 
       for (const defaultConfigFile of defaultConfigFiles) {
-        if (!fs.existsSync(defaultConfigFile.path)) {
+        if (!fs.existsSync(defaultConfigFile)) {
           continue;
         }
 
@@ -1889,16 +1991,16 @@ class WebpackCLI implements IWebpackCLI {
       }
 
       if (foundDefaultConfigFile) {
-        const loadedConfig = await loadConfigByPath(foundDefaultConfigFile.path, options.argv);
+        const loadedConfig = await loadConfigByPath(foundDefaultConfigFile, options.argv);
 
         config.options = loadedConfig.options as WebpackConfiguration[];
 
         if (Array.isArray(config.options)) {
-          config.options.forEach((item) => {
-            config.path.set(item, loadedConfig.path);
-          });
+          for (const item of config.options) {
+            config.path.set(item, [loadedConfig.path]);
+          }
         } else {
-          config.path.set(loadedConfig.options, loadedConfig.path);
+          config.path.set(loadedConfig.options, [loadedConfig.path]);
         }
       }
     }
@@ -1932,6 +2034,92 @@ class WebpackCLI implements IWebpackCLI {
       }
     }
 
+    const resolveExtends = async (
+      config: WebpackConfiguration,
+      configPaths: WebpackCLIConfig["path"],
+      extendsPaths: string[],
+    ): Promise<WebpackConfiguration> => {
+      delete config.extends;
+
+      const loadedConfigs = await Promise.all(
+        extendsPaths.map((extendsPath) =>
+          loadConfigByPath(path.resolve(extendsPath), options.argv),
+        ),
+      );
+
+      const merge = await this.tryRequireThenImport<typeof webpackMerge>("webpack-merge");
+      const loadedOptions = loadedConfigs.flatMap((config) => config.options);
+
+      if (loadedOptions.length > 0) {
+        const prevPaths = configPaths.get(config);
+        const loadedPaths = loadedConfigs.flatMap((config) => config.path);
+
+        if (prevPaths) {
+          const intersection = loadedPaths.filter((element) => prevPaths.includes(element));
+
+          if (intersection.length > 0) {
+            this.logger.error(`Recursive configuration detected, exiting.`);
+            process.exit(2);
+          }
+        }
+
+        config = merge(
+          ...(loadedOptions as [WebpackConfiguration, ...WebpackConfiguration[]]),
+          config,
+        );
+
+        if (prevPaths) {
+          configPaths.set(config, [...prevPaths, ...loadedPaths]);
+        }
+      }
+
+      if (config.extends) {
+        const extendsPaths = typeof config.extends === "string" ? [config.extends] : config.extends;
+
+        config = await resolveExtends(config, configPaths, extendsPaths);
+      }
+
+      return config;
+    };
+
+    // The `extends` param in CLI gets priority over extends in config file
+    if (options.extends && options.extends.length > 0) {
+      const extendsPaths = options.extends;
+
+      if (Array.isArray(config.options)) {
+        config.options = await Promise.all(
+          config.options.map((options) => resolveExtends(options, config.path, extendsPaths)),
+        );
+      } else {
+        // load the config from the extends option
+        config.options = await resolveExtends(config.options, config.path, extendsPaths);
+      }
+    }
+    // if no extends option is passed, check if the config file has extends
+    else if (Array.isArray(config.options) && config.options.some((options) => options.extends)) {
+      config.options = await Promise.all(
+        config.options.map((options) => {
+          if (options.extends) {
+            return resolveExtends(
+              options,
+              config.path,
+              typeof options.extends === "string" ? [options.extends] : options.extends,
+            );
+          } else {
+            return options;
+          }
+        }),
+      );
+    } else if (!Array.isArray(config.options) && config.options.extends) {
+      config.options = await resolveExtends(
+        config.options,
+        config.path,
+        typeof config.options.extends === "string"
+          ? [config.options.extends]
+          : config.options.extends,
+      );
+    }
+
     if (options.merge) {
       const merge = await this.tryRequireThenImport<typeof webpackMerge>("webpack-merge");
 
@@ -1949,11 +2137,13 @@ class WebpackCLI implements IWebpackCLI {
         const configPath = config.path.get(options);
         const mergedOptions = merge(accumulator, options);
 
-        mergedConfigPaths.push(configPath as string);
+        if (configPath) {
+          mergedConfigPaths.push(...configPath);
+        }
 
         return mergedOptions;
       }, {});
-      config.path.set(config.options, mergedConfigPaths as unknown as string);
+      config.path.set(config.options, mergedConfigPaths);
     }
 
     return config;
@@ -1963,21 +2153,6 @@ class WebpackCLI implements IWebpackCLI {
     config: WebpackCLIConfig,
     options: Partial<WebpackDevServerOptions>,
   ): Promise<WebpackCLIConfig> {
-    const runFunctionOnEachConfig = (
-      options: ConfigOptions | ConfigOptions[],
-      fn: CallableFunction,
-    ) => {
-      if (Array.isArray(options)) {
-        for (let item of options) {
-          item = fn(item);
-        }
-      } else {
-        options = fn(options);
-      }
-
-      return options;
-    };
-
     if (options.analyze) {
       if (!this.checkPackageExists("webpack-bundle-analyzer")) {
         await this.doInstall("webpack-bundle-analyzer", {
@@ -2003,35 +2178,21 @@ class WebpackCLI implements IWebpackCLI {
 
     const CLIPlugin = await this.tryRequireThenImport<
       Instantiable<CLIPluginClass, [CLIPluginOptions]>
-    >("./plugins/CLIPlugin");
+    >("./plugins/cli-plugin");
 
     const internalBuildConfig = (item: WebpackConfiguration) => {
-      // Output warnings
-      if (
-        item.watch &&
-        options.argv &&
-        options.argv.env &&
-        (options.argv.env["WEBPACK_WATCH"] || options.argv.env["WEBPACK_SERVE"])
-      ) {
-        this.logger.warn(
-          `No need to use the '${
-            options.argv.env["WEBPACK_WATCH"] ? "watch" : "serve"
-          }' command together with '{ watch: true }' configuration, it does not make sense.`,
-        );
-
-        if (options.argv.env["WEBPACK_SERVE"]) {
-          item.watch = false;
-        }
-      }
+      const originalWatchValue = item.watch;
 
       // Apply options
-      const args: Record<string, Argument> = this.getBuiltInOptions()
-        .filter((flag) => flag.group === "core")
-        .reduce((accumulator: Record<string, Argument>, flag) => {
-          accumulator[flag.name] = flag as unknown as Argument;
+      const args: Record<string, Argument> = this.getBuiltInOptions().reduce(
+        (accumulator: Record<string, Argument>, flag) => {
+          if (flag.group === "core") {
+            accumulator[flag.name] = flag as unknown as Argument;
+          }
           return accumulator;
-        }, {});
-
+        },
+        {},
+      );
       const values: ProcessedArguments = Object.keys(options).reduce(
         (accumulator: ProcessedArguments, name) => {
           if (name === "argv") {
@@ -2049,38 +2210,58 @@ class WebpackCLI implements IWebpackCLI {
         {},
       );
 
-      const problems: Problem[] | null = this.webpack.cli.processArguments(args, item, values);
+      if (Object.keys(values).length > 0) {
+        const problems: Problem[] | null = this.webpack.cli.processArguments(args, item, values);
 
-      if (problems) {
-        // eslint-disable-next-line @typescript-eslint/no-explicit-any
-        const groupBy = (xs: Record<string, any>[], key: string) => {
-          return xs.reduce((rv, x) => {
-            (rv[x[key]] = rv[x[key]] || []).push(x);
+        if (problems) {
+          // eslint-disable-next-line @typescript-eslint/no-explicit-any
+          const groupBy = (xs: Record<string, any>[], key: string) => {
+            return xs.reduce((rv, x) => {
+              (rv[x[key]] = rv[x[key]] || []).push(x);
 
-            return rv;
-          }, {});
-        };
-        const problemsByPath = groupBy(problems, "path");
+              return rv;
+            }, {});
+          };
+          const problemsByPath = groupBy(problems, "path");
 
-        for (const path in problemsByPath) {
-          const problems = problemsByPath[path];
+          for (const path in problemsByPath) {
+            const problems = problemsByPath[path];
 
-          problems.forEach((problem: Problem) => {
-            this.logger.error(
-              `${this.capitalizeFirstLetter(problem.type.replace(/-/g, " "))}${
-                problem.value ? ` '${problem.value}'` : ""
-              } for the '--${problem.argument}' option${
-                problem.index ? ` by index '${problem.index}'` : ""
-              }`,
-            );
+            for (const problem of problems) {
+              this.logger.error(
+                `${this.capitalizeFirstLetter(problem.type.replace(/-/g, " "))}${
+                  problem.value ? ` '${problem.value}'` : ""
+                } for the '--${problem.argument}' option${
+                  problem.index ? ` by index '${problem.index}'` : ""
+                }`,
+              );
 
-            if (problem.expected) {
-              this.logger.error(`Expected: '${problem.expected}'`);
+              if (problem.expected) {
+                this.logger.error(`Expected: '${problem.expected}'`);
+              }
             }
-          });
-        }
+          }
 
-        process.exit(2);
+          process.exit(2);
+        }
+      }
+
+      // Output warnings
+      if (
+        options.isWatchingLikeCommand &&
+        options.argv &&
+        options.argv.env &&
+        (typeof originalWatchValue !== "undefined" || typeof options.argv.watch !== "undefined")
+      ) {
+        this.logger.warn(
+          `No need to use the '${
+            options.argv.env["WEBPACK_WATCH"] ? "watch" : "serve"
+          }' command together with '{ watch: true | false }' or '--watch'/'--no-watch' configuration, it does not make sense.`,
+        );
+
+        if (options.argv.env["WEBPACK_SERVE"]) {
+          item.watch = false;
+        }
       }
 
       const isFileSystemCacheOptions = (
@@ -2105,13 +2286,13 @@ class WebpackCLI implements IWebpackCLI {
           }
 
           if (Array.isArray(configPath)) {
-            configPath.forEach((oneOfConfigPath) => {
+            for (const oneOfConfigPath of configPath) {
               (
                 item.cache.buildDependencies as NonNullable<
                   FileSystemCacheOptions["cache"]["buildDependencies"]
                 >
               ).defaultConfig.push(oneOfConfigPath);
-            });
+            }
           } else {
             item.cache.buildDependencies.defaultConfig.push(configPath);
           }
@@ -2167,13 +2348,18 @@ class WebpackCLI implements IWebpackCLI {
           helpfulOutput: !options.json,
           progress: options.progress,
           analyze: options.analyze,
+          isMultiCompiler: Array.isArray(config.options),
         }),
       );
-
-      return options;
     };
 
-    runFunctionOnEachConfig(config.options, internalBuildConfig);
+    if (Array.isArray(config.options)) {
+      for (const item of config.options) {
+        internalBuildConfig(item);
+      }
+    } else {
+      internalBuildConfig(config.options);
+    }
 
     return config;
   }
@@ -2317,6 +2503,7 @@ class WebpackCLI implements IWebpackCLI {
 
     if (isWatchCommand) {
       options.watch = true;
+      options.isWatchingLikeCommand = true;
     }
 
     compiler = await this.createCompiler(options as WebpackDevServerOptions, callback);
